--------------------------------------------------------------------------------
-- Universidad Autonoma de Madrid
-- Escuela Politecnica Superior
-- Laboratorio de DIE
--------------------------------------------------------------------------------
-- Block: switches_leds
--
-- Author: (UAM)
--
-- Description:
--   Peripheral for PicoBlaze providing registers to interface the switches
--   (slide and pushbutton) and LEDs in S3BOARD.
--   If connected directly to this board, the polarity is as follows:
--   A 1 in a switch means "up" position, a 1 in a pushbutton means "pressed",
--   a 1 turns a LED on.
--
--   Configuration registers map (offset address):
--      0 - Switches (SW0 to SW7) values in corresponding bits (read only)
--      1 - Pushbuttons (BTN0 to BTN3) values in corresponding bits (read only)
--             other bits are read as '0'. (Note that some pushbutton may be
--             reserved externally for reset and not provided to this block).
--      2 - LEDs (LD0 to LD7) values in corresponding bits (read/write)
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

--------------------------------------------------------------------------------
-- Entity declaration
--------------------------------------------------------------------------------

entity switches_leds is
  port (
        -- System signals
        Clk       : in  std_logic; -- System clock
        Reset     : in  std_logic; -- System asynchronous reset, active high
        
        -- Configuration Registers interface
        Sel       : in  std_logic;                     -- If 1 block is accessed
        WriteEn   : in  std_logic;                     -- If 1 access is write
        Address   : in  std_logic_vector (1 downto 0); -- Offset address of reg.
        WData     : in  std_logic_vector (7 downto 0); -- Write data
        RData     : out std_logic_vector (7 downto 0); -- Read data
        
        -- External Input/Output interface:
        SlideSwitches : in  std_logic_vector (7 downto 0); -- Switches
        PushButtons   : in  std_logic_vector (3 downto 0); -- Pushbuttons
        Leds          : out std_logic_vector (7 downto 0)  -- LEDs
       );
end switches_leds;

--------------------------------------------------------------------------------
-- Architecture
--------------------------------------------------------------------------------

architecture rtl of switches_leds is

  --  Constants defining register addresses:
  constant ADDR_SWITCHES    : std_logic_vector (1 downto 0) := "00";
  constant ADDR_PUSHBUTTONS : std_logic_vector (1 downto 0) := "01";
  constant ADDR_LEDS        : std_logic_vector (1 downto 0) := "10";
  
  -- Signal declarations:
  
  --   Anti-metastability chains for switches and pushbuttons:
  signal slideSwMeta0 : std_logic_vector (7 downto 0); -- first FF in chain
  signal slideSwMeta1 : std_logic_vector (7 downto 0); -- second FF in chain
  signal pushButMeta0 : std_logic_vector (3 downto 0); -- first FF in chain
  signal pushButMeta1 : std_logic_vector (3 downto 0); -- second FF in chain
  --   LED value registers:
  signal ledsValue    : std_logic_vector (7 downto 0);
  
begin

  -- Anti-metastability chains for inputs
  process (Clk, Reset)
  begin
    if Reset = '1' then
      slideSwMeta0 <= (others => '0');
      slideSwMeta1 <= (others => '0');
      pushButMeta0 <= (others => '0');
      pushButMeta1 <= (others => '0');
    elsif Clk'event and Clk = '1' then
      slideSwMeta0 <= SlideSwitches;
      slideSwMeta1 <= slideSwMeta0;
      pushButMeta0 <= PushButtons;
      pushButMeta1 <= pushButMeta0;
    end if;
  end process;  
  
  -- Write LEDs:
  process (Clk, Reset)
  begin
    if Reset = '1' then
      ledsValue <= (others => '0');
    elsif Clk'event and Clk = '1' then
      if (Sel = '1') and (WriteEn = '1') and (Address = ADDR_LEDS) then
        ledsValue <= WData;
      end if;
    end if;
  end process;  

  Leds <= ledsValue; -- we use "ledsValue" intermediate signal because we want
                     -- to read it inside this block

  -- Read Slide/Push-Button Switches and programmed LED values:
  process (Address, slideSwMeta1, pushButMeta1, ledsValue)
  begin
    case Address is
      when ADDR_SWITCHES    => RData <= slideSwMeta1;
      when ADDR_PUSHBUTTONS => RData <= "0000" & pushButMeta1;
      when ADDR_LEDS        => RData <= ledsValue;
      when others           => RData <= (others => '0');
    end case;
  end process;    
 
end rtl;


 