-- SRAM 256kx16 Memory model
--
--  F.Barbero - EPS - Universidad Autonoma de Madrid
--
--  Simple memory model fixed for 256k x 16 size. It could be easily 
--  made generic for any depth, but it is optimized for 16-bit width.
--  The model makes use of an integer array to limit the machine memory
--  used by the simulator. The upper bits of each integer word are used
--  to flag 'X' / 'U' conditions for each data bit (in the lower part):
--  if a flag in bit k is 1, then a bit '0' in k-16 means 'X', while
--  a bit '1' in k-16 means 'U'. For writings, any bit different to '0'
--  or '1' is written as 'X'.

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_signed.all;

entity SRAM256kx16 is
   port (
         Ce_n         : in    std_logic;
         Oe_n         : in    std_logic;
         We_n         : in    std_logic;
         Lb_n         : in    std_logic;
         Ub_n         : in    std_logic;
         Addr         : in    std_logic_vector (17 downto 0);
         DataIO       : inout std_logic_vector (15 downto 0)
        );
end SRAM256kx16;

architecture behavioral of SRAM256kx16 is

   constant TAA : time := 10 ns; -- Address access time: while reading,
                                 --   delay from Addr to DataIO
                                 
   type t_mem is array (0 to (2**18-1)) of integer; -- Type for memory
   signal mem : t_mem; -- Memory
   
   signal write    : std_logic;                      -- Internal write signal
   signal dataRead : std_logic_vector (15 downto 0); -- Internal read data
   
   -- Create a 32-bit word with 'X' flags for given 16-bit data, to put that
   -- information in memory:
   function set_x (constant data : in std_logic_vector (15 downto 0))
                            return std_logic_vector is
     variable flagged_word : std_logic_vector (31 downto 0);
   begin
      for i in data'range loop
         if (data(i) /= '0') and (data(i) /= '1') then -- (weak 0/1 considered as unknown)
            -- Mark bit as 'X' (flag='1', data='0')
            flagged_word(i+16) := '1';
            flagged_word(i)    := '0';
         else
            flagged_word(i+16) := '0';
            flagged_word(i)    := data(i);
         end if;
      end loop;
      return flagged_word;
   end function;

   -- Get 'X' / 'U' information from 32-bit memory word and generate the 
   --   corresponding 16-bit data output word:
   function get_x_unknown (constant data : in std_logic_vector (31 downto 0))
                            return std_logic_vector is
     variable result_word : std_logic_vector (15 downto 0);
   begin
      for i in 0 to 15 loop
         if (data(i+16) = '1') and (data(i) = '1') then
            result_word (i) := 'U';
         elsif (data(i+16) = '1') and (data(i) = '0') then
            result_word (i) := 'X';
         else
            result_word (i) := data(i);
         end if;
      end loop;
      return result_word;
   end function;

   signal init : std_logic := '0'; -- auxiliar signal to trigger initialization
   signal update_signal : std_logic := '0'; -- toggles to signal data has been written
   
begin
   
   -- Define write condition: it will happen in the rising edge of the following
   --  signal, 0 (active low) when chip enable and write enable and at least
   --  one byte enable.   
   write <= not ( (not Ce_n) and ((not Lb_n) or (not Ub_n)) and (not We_n) );

   -- Generate a 0->1 transition to initialize the memory in the next process:
   process
   begin
      init <= '1' after 0 ns;
      wait;
   end process;
   
   -- Write process / initialization process:
   write_proc : process (write, init)
     variable current_word  : std_logic_vector (31 downto 0);
     variable new_word      : std_logic_vector (31 downto 0);
   begin
      if rising_edge (init) then
         -- Initialization to unknowns:
         for i in mem'range loop
            mem(i) <= -1; -- All-'1' means all-'U'
         end loop;
      elsif rising_edge (write) then
         -- Write memory integer data with 16-bit input data and 16-bit
         --  corresponding flags:
         current_word  := conv_std_logic_vector (mem (conv_integer(Addr)), 32);
         if (Lb_n = '0') and (Ub_n = '0') then
            new_word := set_x (DataIO);
            mem (conv_integer(Addr)) <= conv_integer(new_word);
         elsif (Lb_n = '0') or (Ub_n = '0') then
            if (Lb_n = '0') then
               new_word := (current_word and x"FF00FF00") or
                            (set_x (DataIO) and x"00FF00FF");
            else -- -> Ub_n = '0'
               new_word := (current_word and x"00FF00FF") or
                            (set_x (DataIO) and x"FF00FF00");
            end if;
            mem (conv_integer(Addr)) <= conv_integer(new_word);
         end if;
         update_signal <= not update_signal;
      end if;
   end process;
   
   -- Read process:
   read_proc : process (Addr, update_signal)
      variable current_data : std_logic_vector (31 downto 0);
   begin
      current_data := conv_std_logic_vector (mem (conv_integer(Addr)),32);
      dataRead <= get_x_unknown (current_data) after TAA;
   end process;
   
   -- Tristate data output buffer:
   DataIO (7  downto 0) <= dataRead (7  downto 0) when (OE_n = '0') and (Lb_n = '0') else (others => 'Z');
   DataIO (15 downto 8) <= dataRead (15 downto 8) when (OE_n = '0') and (Ub_n = '0') else (others => 'Z');
   
end behavioral;
